/***********************************************************
 * $Id: fancyqt4asyncinvocationhandler.h 1424 2015-01-01 12:55:13Z fancylibs $
 *
 * FancyQT4
 *
 * Copyright (C) 2009-2011 ev-i GmbH & IBCL e.U.
 * Copyright (C) 2011-2013 ITEG IT-Engineers GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ************************************************************/

#ifndef FANCYQT4ASYNCINVOCATIONHANDLER_H_
#define FANCYQT4ASYNCINVOCATIONHANDLER_H_

#define FANCYQT4_ASYNCINVOCATIONHANDLER_BEAN "Fancy::Qt4::AsyncInvocationHandler"

#include <fancyqt4.h>
#include <fancyobject.h>

class QObject;

namespace Fancy {

class SessionContextFactory;
class ReflectionTarget;
class ThreadPool;
class ApplicationContext;

namespace Qt4 {

class InvocationResultEmitter;

/**
 * This object is used to dispatch a call to a @c Fancy::ReflectionTarget to
 * a background thread and deliver the result via Qt4's asynchronous signal
 * delivery mechanism.
 *
 * This object will go out of scope after the last asynchronous invocation has finished.
 */
class FANCYQT4_API AsyncInvocationHandler: public virtual Fancy::Object {

public:

    typedef Fancy::ObjPtr<AsyncInvocationHandler> ptr_t;

    static ptr_t newInstance(const Fancy::ObjPtr<Fancy::ThreadPool> &pool,
                             const Fancy::ObjPtr<Fancy::SessionContextFactory> &scf);

    /**
     * Set an ephemeral context to passed to the emitted signals as context parameter.
     * This parameter is set to null when the next asynchronous call is issued by a
     * bean returned by @c wrapTarget or @c wrapBeanTarget.
     *
     * @param ctxt The context object for the next asynchronous call to set.
     */
    virtual void setContext(const Fancy::Object::ptr_t &ctxt) = 0;

    /**
     *
     * @return The ephemeral context object for the next asynchronous call.
     */
    virtual Fancy::Object::ptr_t getContext() const = 0;

    /**
     * Generate a reflection target, which queues an invocation to the underlying thread
     * pool and returns @c null.
     *
     * @param bean The bean name to pass to @c InvocationResult or @c InvocationException.
     * @param delegate The bean to wrap.
     */
    virtual Fancy::ObjPtr<ReflectionTarget> wrapTarget(const std::string &bean,
                                                       const Fancy::ObjPtr<Fancy::ReflectionTarget> &delegate) = 0;

    /**
     * Generate a reflection target, which queues an invocation to the underlying thread
     * pool and returns @c null. The wrapped invocation target is retrieved from an application context.
     *
     * @param bean The bean name to pass to @c InvocationResult or @c InvocationException.
     *             This is also the name of the bean retrieved from the application context.
     * @param ctxt The application context to get the bean to wrap from.
     */
    virtual Fancy::ObjPtr<ReflectionTarget> wrapTarget(const std::string &bean,
                                                       const Fancy::ObjPtr<Fancy::ApplicationContext> &ctxt) = 0;

    /**
     * Generate a reflection target, which queues an invocation to the underlying thread
     * pool and returns @c null. The wrapped invocation target is retrieved from the
     * static application context.
     *
     * @param bean The bean name to pass to @c InvocationResult or @c InvocationException.
     *             This is also the name of the bean retrieved from the static application context.
     */
    virtual Fancy::ObjPtr<ReflectionTarget> wrapStaticTarget(const std::string &bean) = 0;

    /**
     * Generate a bean instance, which queues an invocation to the underlying thread
     * pool and returns @c null.
     *
     * @param Bean The interface class to wrap.
     * @param bean The bean name to pass to @c InvocationResult or @c InvocationException.
     * @param delegate The bean to wrap.
     */
    template<class Bean>
    Fancy::ObjPtr<Bean> wrapBeanTarget(const std::string &bean,
                                       const Fancy::ObjPtr<Fancy::ReflectionTarget> &delegate) {
        return Bean::newProxyInstance(this->wrapTarget(bean,delegate));
    }

    /**
     * Generate a reflection target, which queues an invocation to the underlying thread
     * pool and returns @c null. The wrapped invocation target is retrieved from an application context.
     *
     * @param Bean The interface class to wrap.
     * @param bean The bean name to pass to @c InvocationResult or @c InvocationException.
     *             This is also the name of the bean retrieved from the application context.
     * @param ctxt The application context to get the bean to wrap from.
     */
    template<class Bean>
    Fancy::ObjPtr<Bean> wrapBeanTarget(const std::string &bean,
                                       const Fancy::ObjPtr<Fancy::ApplicationContext> &ctxt) {

        return Bean::newProxyInstance(this->wrapTarget(bean,ctxt));
    }

    /**
     * Generate a reflection target, which queues an invocation to the underlying thread
     * pool and returns @c null. The wrapped invocation target is retrieved from the
     * static application context.
     *
     * @param Bean The interface class to wrap.
     * @param bean The bean name to pass to @c InvocationResult or @c InvocationException.
     *             This is also the name of the bean retrieved from the static application context.
     */
    template<class Bean>
    Fancy::ObjPtr<Bean> wrapStaticBeanTarget(const std::string &bean) {

        return Bean::newProxyInstance(this->wrapStaticTarget(bean));
    }

    /**
     * @return The QObject, which's signals are emitted upon the finalization of the
     *         dispatched operation.
     */
    virtual InvocationResultEmitter *getEmitter() const = 0;

    /**
     * A shortcut for
     * <pre>
     *  QObject::connect(this->getEmitter(),
     *                   SIGNAL(resultReceived(const Fancy::Qt4::InvocationResult::ptr_t&)),
     *                   receiver,method);
     * </pre>
     *
     * Please note, that due to limitations of Qt's asynchronous signal/slot handling, you should
     * always include @c fancyqt4invocationresult.h in your header files and use the
     * @c Fancy::Qt4::InvocationException::ptr_t notation instead of the notation
     * @c Fancy::ObjPtr<Fancy::Qt4::InvocationException>.
     *
     * @param receiver The receiving Object, in most cases a GUI widget.
     * @param qt_slot The receiving slot which should be constructed by the @c SLOT macro.
     * @return the return value of QObject::connect()
     */
    virtual bool connectResultReceived(QObject *receiver, const char *qt_slot) const = 0;

    /**
     * A shortcut for
     * <pre>
     *  QObject::connect(this->getEmitter(),
     *                   SIGNAL(exceptionReceived(const Fancy::Qt4::InvocationException::ptr_t&)),
     *                   receiver,method);
     * </pre>
     *
     * Please note, that due to limitations of Qt's asynchronous signal/slot handling, you should
     * always include @c fancyqt4invocationresult.h in your header files and use the
     * @c Fancy::Qt4::InvocationException::ptr_t notation instead of the notation
     * @c Fancy::ObjPtr<Fancy::Qt4::InvocationException>.
     *
     * @param receiver The receiving Object, in most cases a GUI widget.
     * @param qt_slot The receiving slot which should be constructed by the @c SLOT macro.
     * @return the return value of QObject::connect()
     */
    virtual bool connectExceptionReceived(QObject *receiver, const char * qt_slot) const = 0;

    /**
     * @param methodName The name of an RPC method for which a dedicated
     *         emitter is returned.
     * @return A QObject, which's signals are emitted upon the finalization of the
     *         dispatched operation for the given method name.
     */
    virtual InvocationResultEmitter *getEmitter(const std::string &methodName) const = 0;

    /**
     * A shortcut for
     * <pre>
     *  QObject::connect(this->getEmitter(methodName),
     *                   SIGNAL(resultReceived(const Fancy::Qt4::InvocationResult::ptr_t&)),
     *                   receiver,method);
     * </pre>
     *
     * Please note, that due to limitations of Qt's asynchronous signal/slot handling, you should
     * always include @c fancyqt4invocationresult.h in your header files and use the
     * @c Fancy::Qt4::InvocationException::ptr_t notation instead of the notation
     * @c Fancy::ObjPtr<Fancy::Qt4::InvocationException>.
     *
     * @param methodName The name of an RPC method for which a dedicated
     *         emitter is used.
     * @param receiver The receiving Object, in most cases a GUI widget.
     * @param qt_slot The receiving slot which should be constructed by the @c SLOT macro.
     * @return the return value of QObject::connect()
     */
    virtual bool connectResultReceived(const std::string &methodName, QObject *receiver, const char *qt_slot) const = 0;

    /**
     * A shortcut for
     * <pre>
     *  QObject::connect(this->getEmitter(methodName),
     *                   SIGNAL(exceptionReceived(const Fancy::Qt4::InvocationException::ptr_t&)),
     *                   receiver,method);
     * </pre>
     *
     * Please note, that due to limitations of Qt's asynchronous signal/slot handling, you should
     * always include @c fancyqt4invocationresult.h in your header files and use the
     * @c Fancy::Qt4::InvocationException::ptr_t notation instead of the notation
     * @c Fancy::ObjPtr<Fancy::Qt4::InvocationException>.
     *
     * @param methodName The name of an RPC method for which a dedicated
     *         emitter is used.
     * @param receiver The receiving Object, in most cases a GUI widget.
     * @param qt_slot The receiving slot which should be constructed by the @c SLOT macro.
     * @return the return value of QObject::connect()
     */
    virtual bool connectExceptionReceived(const std::string &methodName, QObject *receiver, const char * qt_slot) const = 0;

protected:
    AsyncInvocationHandler();
    virtual ~AsyncInvocationHandler();
};

}}

#endif /* FANCYQT4ASYNCINVOCATIONHANDLER_H_ */
